import pandas as pd
from datetime import datetime, timedelta
from collections import Counter, defaultdict

from smappPy.retweet import is_retweet
from smappPy.geo_tweet import is_geocoded
from smappPy.entities import contains_url, contains_image, contains_hashtag, contains_mention

import mongo_tweet_collection
from counter_functions import _top_user_locations, _top_unigrams, _top_bigrams, _top_trigrams, _top_links, _top_urls, \
    _top_images, _top_hashtags, _top_mentions, _top_geolocation_names, _counter_to_series, _language_counts, \
    _unique_users


class Aggregator(object):
    """
    Aggregator class used to produce aggregate results grouped by time slice.
    Time slice can be 'days', 'hours', 'minutes', or 'seconds'. Iterates over collection, splitting by that time unit,
    and produces results for each of those splits.

    Usage example:
    ##############
        col = smapp_toolkit.twitter.BSONTweetCollection('tweets.bson') # or MongoTweetCollection()
        agg = smapp_toolkit.twitter.aggregator.Aggregator(col, 'hours')
        d = agg.top_user_locations(10)
    """

    def __init__(self, collection, time_unit):
        self._collection = collection
        self._time_unit = time_unit
        self._time_delta = timedelta(**{time_unit: 1})

    def _get_start_time(self):
        for t in self._collection:
            start_time = t['timestamp']
            break
        if self._time_unit == 'days':
            start_time = start_time.replace(hour=0, minute=0, second=0, microsecond=0)
        elif self._time_unit == 'hours':
            start_time = start_time.replace(minute=0, second=0, microsecond=0)
        elif self._time_unit == 'minutes':
            start_time = start_time.replace(second=0, microsecond=0)
        elif self._time_unit == 'seconds':
            start_time = start_time.replace(microsecond=0)
        return start_time

    def __iter__(self):
        if isinstance(self._collection, mongo_tweet_collection.MongoTweetCollection):
            return self._mongo_splits()
        else:
            return self._splits()

    def _splits(self):
        start_time = self._get_start_time()

        it = iter(self._collection)
        d = dict()
        d['t'] = next(it)
        d['stop'] = False

        def chunk(until):
            while d['t']['timestamp'] < until:
                yield d['t']
                try:
                    d['t'] = next(it)
                except StopIteration:
                    d['stop'] = True
                    raise StopIteration()
            raise StopIteration()

        while True:
            ch = chunk(start_time + self._time_delta)
            yield (start_time, ch)
            if d['stop']:
                raise StopIteration()
            start_time = start_time + self._time_delta

    def _mongo_splits(self):
        if not self._collection._get_until():
            raise Exception("Cannot time-split on a MongoTweetCollection without calling until()")
        start_time = self._get_start_time()
        end_time = self._collection._get_until()

        while start_time < end_time:
            tmpcol = self._collection._copy()
            tmpcol._override_since(start_time)
            tmpcol._override_until(start_time+self._time_delta)
            yield (start_time, tmpcol)
            start_time = start_time + self._time_delta

    def grouped_result(self, callable_, *args, **kwargs):
        results = dict()
        for t, split in self:
            results[t] = callable_(split, *args, **kwargs)
        return pd.concat(results, axis=1).T

    def grouped_top_n_result(self, n, callable_):
        res = self.grouped_result(callable_)
        su = res.sum()
        su.sort(ascending=False)
        return res[list(su[:n].index)]

    def top_user_locations(self, n=10):
        return self.grouped_top_n_result(n, _top_user_locations)

    def top_unigrams(self, n=10, hashtags=True, mentions=True, rts=False, mts=False, https=False, stopwords=[]):
        return self.grouped_top_n_result(n, lambda col: _top_unigrams(col, n=None, hashtags=hashtags, mentions=mentions, rts=rts, mts=mts, https=https, stopwords=stopwords))

    def top_bigrams(self, n=10, hashtags=True, mentions=True, rts=False, mts=False, https=False, stopwords=[]):
        return self.grouped_top_n_result(n, lambda col: _top_bigramss(col, n=None, hashtags=hashtags, mentions=mentions, rts=rts, mts=mts, https=https, stopwords=stopwords))

    def top_trigrams(self, n=10, hashtags=True, mentions=True, rts=False, mts=False, https=False, stopwords=[]):
        return self.grouped_top_n_result(n, lambda col: _top_trigrams(col, n=None, hashtags=hashtags, mentions=mentions, rts=rts, mts=mts, https=https, stopwords=stopwords))

    def top_links(self, n=10):
        return self.grouped_top_n_result(n, _top_links)

    def top_urls(self, n=10):
        return self.grouped_top_n_result(n, _top_urls)

    def top_images(self, n=10):
        return self.grouped_top_n_result(n, _top_images)

    def top_hashtags(self, n=10):
        return self.grouped_top_n_result(n, _top_hashtags)

    def top_mentions(self, n=10):
        return self.grouped_top_n_result(n, _top_mentions)

    def top_geolocation_names(self, n=10):
        return self.grouped_top_n_result(n, _top_geolocation_names)

    def entities_counts(self, urls=True, images=True, hashtags=True, mentions=True, geo_enabled=True, retweets=True):
        def props(collection, urls=urls, images=images, hashtags=hashtags, mentions=mentions, geo_enabled=geo_enabled, retweets=retweets):
            res = Counter()
            for tweet in collection:
                res['_total'] += 1
                if urls and contains_url(tweet):
                    res['url'] += 1
                if images and contains_image(tweet):
                    res['image'] += 1
                if hashtags and contains_hashtag(tweet):
                    res['hashtag'] += 1
                if mentions and contains_mention(tweet):
                    res['mention'] += 1
                if geo_enabled and is_geocoded(tweet):
                    res['geo_enabled'] += 1
                if retweets and is_retweet(tweet):
                    res['retweet'] += 1
            return _counter_to_series(res)
        return self.grouped_result(props)

    def language_counts(self, langs):
        return self.grouped_result(_language_counts, langs=langs)

    def count(self):
        if isinstance(self._collection, mongo_tweet_collection.MongoTweetCollection):
            return self.grouped_result(lambda it: pd.Series(it.count(), index=['count']))
        else:
            return self.grouped_result(lambda it: pd.Series(sum(1 for e in it), index=['count']))

    def unique_users(self):
        return self.grouped_result(_unique_users)



"""
Module for plotting functions that work on data, not on live database connections.

@jonathanronen 2015/6
"""

import pytz
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from collections import OrderedDict

def stacked_bar_plot(data, columns, x_tick_date_format='%Y-%m-%d', x_tick_step=2, bar_width=.8, alpha=.6, colors=sns.color_palette()):
    """
    Makes a stacked bars plot from data in a pandas.DataFrame.
    columns are the columns to be stacked.

    Example: Plot language proportions
    ##################################
        data = collection.group_by('days').language_counts(langs=['en','es','other'])

        plt.figure(figsize=(10,10))
        stacked_bar_plot(data, ['en','es','other'], colors=['royalblue', 'yellow', 'grey'])
        plt.title('Tweet proportions in English and Spanish', fontsize=24)
        plt.tight_layout()

    -----------------------------------------------------------------

    Example: Plot retweet proportion
    ################################
        data = col.since(datetime(2015,6,18,12)).until(datetime(2015,6,18,12,10)).group_by('minutes').entities_counts()
        data['original tweet'] = data['_total'] - data['retweet']

        plt.figure(figsize=(10,10))
        stacked_bar_plot(data, ['retweet', 'original tweet'], colors=['salmon', 'lightgrey'])
        plt.title('Retweet proportion', fontsize=24)
        plt.tight_layout()
    """
    bars = OrderedDict()
    bars[columns[0]] = plt.bar(range(len(data)),
                               data[columns[0]],
                               width=bar_width,
                               linewidth=0.0,
                               color=colors[0],
                               alpha=alpha,
                               label=columns[0])

    for l in columns[1:]:
        bars[l] = plt.bar(range(len(data)),
                          data[l],
                          width=bar_width,
                          linewidth=0.0,
                          color=colors[columns.index(l)],
                          alpha=alpha,
                          bottom=[c.get_y() + c.get_height() for c in bars[columns[columns.index(l)-1]].get_children()],
                          label=l)
    plt.xlim(0, len(data))
    plt.tick_params(axis="x", which="both", bottom="on", top="off", length=8, width=1, color="#999999")
    plt.legend(fontsize=14, loc=0)
    plt.xticks(np.arange(len(data))[::x_tick_step]+.4,
               [x.strftime(x_tick_date_format) for x in data.index[::x_tick_step]],
               rotation=45,
               fontsize=18)
    plt.yticks(fontsize=18)
    plt.tight_layout()

def line_with_annotations(data, events=[], x=None, y='count', x_tick_date_format='%Y-%m-%d', x_tick_timezone='UTC', x_tick_step=2, linewidth=2.0, alpha=.3, line_color='red', x_label_step=2):
    """
    Used to plot, for instance, tweets per day, with vertical lines to annotate events.
      `data` is a pandas.DataFrame object holding the line to be plotted
      `events` is a list of tuples with times, texts and alignments for events to be annotated onto figure
      `x` is the name of the column in the dataframe to use as the x-axis (defaults to the index)
      `y` is the column to plot on the y-axis


    Example (tweets per hour with annotations)
    #########################################
    data = collection.group_by('hours')
    events = [
      (datetime(2015,6,21,10), 'Sunrise', 'top'),
      (datetime(2015,6,21,22), 'Sunset', 'bottom')
    ]

    plt.figure(10,6)
    line_with_annotations(data, events, x_tick_timezone='America/New_York')
    plt.title('Tweets mentioning Donald Trump\non 2015-6-21', fontsize=24)
    """
    x = data[x] if x else data.index
    y = data[y]

    plt.plot(x, y, alpha=alpha, linewidth=linewidth, color=line_color)

    ymin, ymax = plt.ylim()
    for e in events:
        plt.axvline(e[0], linestyle="--", color="#999999")
        if e[2] == "bottom":
            plt.text(e[0], ymin + (0.05 * ymax), e[1], rotation=-90, verticalalignment="bottom")
        else:
            plt.text(e[0], ymax - (0.05 * ymax), e[1], rotation=-90, verticalalignment="top")

    plt.ylabel("# Tweets", fontsize=22)
    plt.tick_params(axis="x", which="both", bottom="on", top="off", length=8, width=1, color="#999999")
    plt.xticks(x[::x_label_step], [e.replace(tzinfo=pytz.utc).astimezone(pytz.timezone(x_tick_timezone)).strftime(x_tick_date_format) for e in x[::x_label_step]], fontsize=16, rotation=90)
    plt.yticks(fontsize=16)
    plt.xlabel('Time\nin {}'.format(x_tick_timezone), fontsize=22)



"""
Module contains helper functions for plotting.

@jonathanronen 4/2015
"""
try:
    import numpy as np
except:
    warnings.warn("Cannot import numpy. Plotting will not work.")
import warnings
try:
    import seaborn as sns
    import matplotlib.pyplot as plt
except:
    warnings.warn("Error importing plotting libraries (seaborn and matplotlib). Plotting functionality will not work.")
from datetime import datetime, timedelta


def plot_histo(d, *args, **kwargs):
    count_by = kwargs.pop('count_by', 'minutes')
    key_format = kwargs.pop('key_format', '%Y-%m-%d %H:%M')

    start_time = datetime.strptime(sorted(d.keys())[0], key_format)
    end_time = datetime.strptime(sorted(d.keys())[-1], key_format)
    if count_by == 'minutes':
        t,y = zip(*[(start_time + i*timedelta(minutes=1), d.get((start_time + i*timedelta(minutes=1)).strftime(key_format), 0)) for i in range(int(np.ceil((end_time - start_time).total_seconds()/60)))])
    elif count_by == 'hours':
        t,y = zip(*[(start_time + i*timedelta(hours=1), d.get((start_time + i*timedelta(hours=1)).strftime(key_format), 0)) for i in range(int(np.ceil((end_time - start_time).total_seconds()/(60*60))))])
    elif count_by == 'days':
        t,y = zip(*[(start_time + i*timedelta(days=1), d.get((start_time + i*timedelta(days=1)).strftime(key_format), 0)) for i in range(int(np.ceil((end_time - start_time).total_seconds()/(60*60*24))))])
    else:
        raise Exception("Can't plot histogram by {}. Legal values are ['minutes', 'hours', 'days'].".format(count_by))
    plt.plot(t,y, *args, **kwargs)
    return t,y


def term_counts_histogram(data, key_format, count_by, plot_total=True):
    """
    Function to make histogram plot for data created using `term_counts()`.

    If the data has the format:
    {
        '2015-01-01 18:10': {
            'justin': 12,
            'miley': 33
        },
        '2015-01-01 18:11': {
            'justin': 11,
            'miley': 9
        }
    }

    Then to make the plot, call

    figure_helpers.term_counts_histogram(data, '%Y-%m-%d %H:%M', count_by='minutes')
    ------------------------------------------------

    Legal values for count_by are ['days', 'hours', 'minutes']
    and the `key_format` is the strftime string for the keys of the data dict.
    """
    colors = sns.color_palette('hls', len(data[data.keys()[0]].keys()))

    terms = data[data.keys()[0]].keys()
    terms.remove('_total')
    for c, term in zip(colors,terms):
        t,y = plot_histo({k : data[k][term] for k in data}, label=term, color=c, count_by=count_by, key_format=key_format)

    if plot_total:
        plot_histo({k: data[k]['_total'] for k in data}, label='total', color='grey', linestyle='--', count_by=count_by, key_format=key_format)

    plt.legend()
    plt.xticks(t[::len(t)/10],
               [ts.strftime(key_format) for ts in t[::len(t)/10]],
               rotation=45)
    plt.tight_layout()
